// Roxen LogView Bucket Classes
//
// Copyright (C) 1996, 1997, 1998 Idonex
//

#define BUCKET_DEBUG 0
#define BUCKET_UNKNOWN_FIELD 0
#define MAGIC 0x42767781
#define SEARCH_SIZE (4096)

#define BUCKET_ERR(msg) throw(tex(({ "(Bucket) "+msg+"\n", backtrace() })))
#define HASH(s) (hash(s) & 0xffffffff)

#if constant(thread_create)
#define THREAD_SAFE
#define LOCK() do { object key; catch(key=lock())
#define UNLOCK() key=0; } while(0)
#else
#undef THREAD_SAFE
#define LOCK() do {
#define UNLOCK() } while(0)
#endif

class bucket {
#ifdef THREAD_SAFE
  static inherit Thread.Mutex;
#endif

  static private string filename;
  static private array cache = ({});
  static private int modify = 0, cache_dirty = 0, cache_size = 1000;

  static private mapping index = ([]);

  static private string current_name;
  static private array current_fields;
  static private int index_loaded = 0;

  static private function warning;

  /*
  ** Exception handling mechanism
  */

  static private int exception_counter = 0;
  static private mixed tex(mixed bt)
  {
    exception_counter++;
    if(warning && exception_counter < 10)
      if(stringp(bt))
	warning("(Bucket) <b>"+bt+"</b>");
      else
	warning(bt);
    return bt;
  } 
  
  /*
  ** Escape routines. "\177" is escaped.
  */

  static private string escape(string s)
  {
    return replace(s, ({ "\177", "\200" }), ({ "\200\200", "\200\201" }));
  }

  static private string descape(string s)
  {
    return replace(s, ({ "\200\200", "\200\201" }), ({ "\177", "\200" }));
  }

  /*
  ** Temporary file management.
  */

  static private array|int read_q(int|void header_only)
    // The cache does not need to be flushed.
  {
    array table = ({});
    int size, checksum;
    string chunks = Stdio.read_bytes(filename+".q");
    if(chunks) {
      foreach(chunks/"\177", string chunk) {
	if(sscanf(descape(chunk), "%4c%4c%s", size, checksum, chunk) == 3 &&
	   size == sizeof(chunk) && HASH(chunk) == checksum) catch {
	  mixed d = decode_value(chunk);
	  if(mappingp(d)) {
	    current_name = d->name;
	    current_fields = d->fields;
	    if(header_only)
	      return 1;
	  } else if(arrayp(d))
	    table += d;
	  else
	    tex("Illegal chunk in read_q.");
	};
      }
    }
    if(header_only || !current_name || !current_fields)
      return 0;
    table += cache;
    return table;
  }

  static private int write_q(mixed d)
    // Write cache to disk.
  {
    int r = 0;
    if(!modify)
      BUCKET_ERR("write_q NA for log '"+filename+"' in read mode");
    
    object file = Stdio.File();
    if(file->open(filename+".q", "wac"))  {
      string s = encode_value(d);
      s = sprintf("%4c%4c%s", sizeof(s), HASH(s), s);
      s = "\177"+escape(s);
      if(!(r = (file->write(s) == sizeof(s))))
	tex("write_q failed to write file. Disk full?");
    } else
      tex("write_q cannot open file. File permission problems?");
    file->close();
    return r;
  }

  /*
  ** Paranoid IO functions
  */

  static private string paranoid_read(object file, int n)
  {
    string s = file->read(n);
    if(sizeof(s) != n)
      BUCKET_ERR("Log '"+filename+"' too short");
    return s;
  }

  static private int paranoid_seek(object file, int n)
  {
    int r = file->seek(n);
    if(r == -1)
      BUCKET_ERR("Cannot seek in log '"+filename+"'");
    return r;
  }

  /*
  ** Log encoding style
  */

  static private string encode_block(mixed m)
  {
#if constant(Gz.deflate)
    string s = escape(Gz.deflate()->deflate(encode_value(m)));
#else
    string s = escape(encode_value(m));
#endif
    return sprintf("%08x%08x%s", sizeof(s), HASH(s), s);
  }

  static private mixed decode_block(object file, int|void offset)
  {
    int size, checksum;
    if(query_num_arg() == 2)
      paranoid_seek(file, offset);
    string s = paranoid_read(file, 16);
    sscanf(s[0..7], "%x", size);
    sscanf(s[8..15], "%x", checksum);
    s = paranoid_read(file, size);
    if(HASH(s) != checksum)
      BUCKET_ERR("Checksum error in log '"+filename+"'");

    s = descape(s);
#if constant(Gz.inflate)
    catch { return decode_value(Gz.inflate()->inflate(s)); };
#endif
    return decode_value(s);
  }

  /*
  ** Index management
  */

  // Handle name conflicts (append a unique #)
  static private void name_conflict(mapping e, mapping m)
  {
    string name = indices(e)[0];
    m[name+sprintf("#%08x", sizeof(glob(name+"#*",indices(m))))] = e[name];
  }

  static private mapping read_index()
  {
    mapping m = ([]);
    foreach((Stdio.read_bytes(filename+".i")||"")/"\177", string s) {
      int size, checksum;
      if(sscanf(s[0..7], "%x", size) && sscanf(s[8..15], "%x", checksum)) {
	s = s[16..];
	if(sizeof(s) == size && HASH(s) == checksum) {
	  catch {
	    mapping entry = decode_value(descape(s));
	    if(!zero_type(m[indices(entry)[0]]))
	      name_conflict(entry, m);
	    else
	      m += entry;
	  };
	}
      }
    }
    return index = m;
  }

  static private int write_index(string name, int offset, string filename)
  {
    object file = Stdio.File();
    if(!file->open(filename+".i", "wac"))
      return 0;
    string s = escape(encode_value(([ name:offset ])));
    int r = file->write(sprintf("\177%08x%08x%s", sizeof(s), HASH(s), s));
    file->close();
    return r == 1 + 8 + 8 + sizeof(s);
  }

  /*
  ** Internal functions
  */
  
  static private int _bundle(string|void bundle_file)
  {
    array table = read_q();
    cache = ({});
    cache_dirty = 0;
    rm(filename+".q");
    if(!table)
      return 0;

    // Go to query mode
    modify = 0;
    index_loaded = 0;

    // Encode table column wise
    int offset = 0;
    string chunked = "";
    array offsets = ({});
    for(int i = 0; i < sizeof(current_fields); i++) {
      string g = encode_block(column(table, i));
      offsets += ({ offset });
      offset += sizeof(g);
      chunked += g;
    }

    mapping chunks = mkmapping(Array.map(current_fields,lower_case), offsets);
    mapping m = ([ "name":current_name, "chunks":chunks, "tablesize":sizeof(table) ]);
    string s = "\177" + encode_block(m) + chunked;

    // Write file
    bundle_file = bundle_file||filename;
    object file = Stdio.File();
    if(!file->open(bundle_file, "wac")) {
      tex("bundle cannot open file. File permission problems?");
      return 0;
    }
    int r = file->write(s);
    if(r == sizeof(s))
      write_index(current_name, file->tell() - r, bundle_file);
    else
      tex("bundle failed to write file. Disk full?");
    file->close();
    return r == sizeof(s);
  }

  static private array transpose(array table)
  {
    array t = ({});
    if(!sizeof(table))
      return ({});
    for(int i = 0; i < sizeof(table[0]); i++)
      t += ({ column(table, i) });
    return t;
  }

  static private mapping read_index_once()
  {
    if(!index_loaded) {
      index_loaded = 1;
      read_index();
    }
    return index;
  }

  static private array _query(string name, array(string) request,
			      array(string)|void sum)
  {
    read_index_once();

    if(zero_type(index[name]))
      BUCKET_ERR("'"+name+"' not in log '"+filename+"'\n");
    int offset = index[name];

    // Read header block
    object file = Stdio.File();
    if(!file->open(filename, "r"))
      BUCKET_ERR("Could not open log '"+filename+"'\n");
    paranoid_seek(file, offset);
    if(paranoid_read(file, 1) != "\177")
      BUCKET_ERR("No magic in log '"+filename+"'\n");
    mapping m = decode_block(file);
    
    // Fetch requested fields
    array table = ({});
    offset = file->tell();
    foreach(request, string field) {
      if(field == "hits")
	table += ({ Array.map(allocate(m->tablesize), lambda(){return 1;}) });
      else if(zero_type(m->chunks[field]))
#if BUCKET_UNKNOWN_FIELD
	BUCKET_ERR("Unknown field '"+field+"' in log '"+filename+"'\n");
#else
        return ({});
#endif
      else
      	table += ({ decode_block(file, offset+m->chunks[field]) });
    }

    table = transpose(table);
#if constant(ArrayOp)
    if(sum)
      table = ArrayOp.array_sum(table,
		Array.map(sum, lambda(string s, mapping m)
			       { return m[s]; },
			  mkmapping(request, indices(request))));
#endif
    return table;
  }

  static private mixed fetch(string|int name_or_offset, string|void request)
  {
    int offset;
    if(stringp(name_or_offset)) {
      string name = name_or_offset;
      read_index_once();
      if(zero_type(index[name]))
	BUCKET_ERR("'"+name+"' not in log '"+filename+"'\n");
      offset = index[name];
    } else
      offset = name_or_offset;

    // Read header block
    object file = Stdio.File();
    if(!file->open(filename, "r"))
      BUCKET_ERR("Could not open log '"+filename+"'\n");
    paranoid_seek(file, offset);
    if(paranoid_read(file, 1) != "\177")
      BUCKET_ERR("No magic in log '"+filename+"'\n");
    mapping m = decode_block(file);

    switch(request) {
    case "fields":
      array fields = indices(m->chunks);
      sort(values(m->chunks), fields);
      return fields;
    }
    return m;
  }

  /*
  ** Interface functions
  */

  int flush()
  {
    LOCK();
    if(modify && cache_dirty) {
      cache_dirty = 0;
      array local_cache = cache;
      cache = cache[sizeof(local_cache)..];
      return write_q(local_cache);
    }
    return 0;
    UNLOCK();
  }

  void add(array row)
  {
    if(!modify)
      BUCKET_ERR("add NA for log '"+filename+"' in read mode");

    cache_dirty = 1;
    cache += ({ copy_value(row) });
    if(sizeof(cache) >= cache_size)
      flush();
  }

  array _indices()
  {
    LOCK();
    return sort(indices(read_index_once()));
    UNLOCK();
  }

  int bundle(mixed ... args)
  {
    LOCK();
    if(!modify)
      BUCKET_ERR("bundle NA for log '"+filename+"' in read mode");
#if BUCKET_DEBUG
    return _bundle(@args);
#else
    int r = 0;
    catch(r = _bundle(@args));
    return r;
#endif
    UNLOCK();
  }

  array query(mixed ... args)
  {
    LOCK();
    if(modify)
      BUCKET_ERR("query NA for log '"+filename+"' in write mode");
#if BUCKET_DEBUG
    return _query(@args);
#else
    mixed m;
    if(catch(m = _query(@args)))
      return ({});
    return m;
#endif
    UNLOCK();
  }

  string name()
  {
    return current_name;
  }
  
  array fields(string name)
  {
    LOCK();
    if(modify)
      BUCKET_ERR("fields NA for log '"+filename+"' in write mode");
#if BUCKET_DEBUG
    return fetch(name, "fields");
#else
    array fields;
    if(catch(fields = fetch(name, "fields")))
      return ({});
    return fields;
#endif
    UNLOCK();
  }
  
  string ditched(string name)
  {
    LOCK();
    if(modify)
      BUCKET_ERR("ditched NA for log '"+filename+"' in write mode");
    mixed err;
    if(err = catch(_query(name, fetch(name, "fields"))))
      return err[0]-"\n";
    return 0;
    UNLOCK();
  }

  array offsets()
  {
    read_index_once();

    mapping m = ([]);
    object file = Stdio.File();
    if(file->open(filename, "r")) {
      string s;
      int offset = 0;
      while((s = file->read(SEARCH_SIZE)) && sizeof(s)) {
	int x = 0;
	while((x = search(s, "\177", x)) != -1) {
	  m[sprintf("DEBUG-%08d", offset+x)] = offset+x;
	  x++;
	}
	offset += SEARCH_SIZE;
      }
      file->close();
    }
    array a = values(index);
    index = m;
    return sort(a);
  }

  void exception(function warning_in)
  {
    warning = warning_in;
  }

  void purge_q()
  {
    rm(filename+".q");
  }
  
  void create(string filename_in, string|void mode,
	      string|void name_in, array(string)|void fields_in)
  {
    filename = filename_in;
    mode = mode||"r";
    modify = sizeof(mode-"w") != sizeof(mode);
    current_name = name_in;
    current_fields = fields_in;
    cache = ({});
    cache_dirty = 0;

    if(modify) {
      if(current_name && current_fields) {
	write_q(([ "name":current_name, "fields":current_fields ]));
      } else {
	if(!read_q(1) || !current_name || !current_fields)
	  BUCKET_ERR("Log '"+filename+"' is not open");
      }
    }
  }

  void destroy()
  {
    flush();
  }
}

#define MAX_ROWS 10000

class log {
#ifdef THREAD_SAFE
  static inherit Thread.Mutex;
#endif

  static private array fields;
  static private object b;
  static private string logfile, hostname, ctx;
  static private int hour, minute, rows = 0;

  static private constant time_fields = ({
    "year", "month", "date", "hour", "minute", "second",
    "host-writing-log", "context" });

  static private array timestamp()
  {
    mapping t = localtime(time());
    return ({ t->year+1900, t->mon+1, t->mday, t->hour, t->min, t->sec });
  }

  void flush()
  {
    if(b)
      b->flush();
  }

  void add(array row)
  {
    LOCK();
    array now = timestamp();
    row = copy_value(row+now+({ hostname, ctx }));

    if(b && minute != now[4])
      b->flush();
    minute = now[4];

    if(b && (hour != now[3] || rows > MAX_ROWS)) {
      b->bundle();
      b = 0;
      rows = 0;
    }
    hour = now[3];

    if(!b) {
      string tablename = sprintf("%04d_%02d_%02d:%02d",@now[0..3]);
      b = bucket(logfile, "w", tablename, fields+time_fields);
    }

    b->add(row);
    rows++;
    UNLOCK();
  }

  void create(string logfile_in, array fields_in, string ctx_in)
  {
    logfile = logfile_in;
    fields = fields_in;
    ctx = ctx_in;
    hostname = gethostname();

    if(!catch(b = bucket(logfile, "w")))
      b->bundle();
    b = 0;
  }
}
